"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const error_1 = require("@pnpm/error");
const logger_1 = require("@pnpm/logger");
const getCredentialsByURI = require("credentials-by-uri");
const mem = require("mem");
const fs = require("mz/fs");
const path = require("path");
const ssri = require("ssri");
const createDownloader_1 = require("./createDownloader");
function default_1(opts) {
    const download = createDownloader_1.default({
        alwaysAuth: opts.alwaysAuth || false,
        ca: opts.ca,
        cert: opts.cert,
        key: opts.key,
        localAddress: opts.localAddress,
        proxy: opts.httpsProxy || opts.proxy,
        registry: opts.registry,
        retry: {
            factor: opts.fetchRetryFactor,
            maxTimeout: opts.fetchRetryMaxtimeout,
            minTimeout: opts.fetchRetryMintimeout,
            retries: opts.fetchRetries,
        },
        // TODO: cover with tests this option
        // https://github.com/pnpm/pnpm/issues/1062
        strictSSL: typeof opts.strictSsl === 'boolean'
            ? opts.strictSsl
            : true,
        userAgent: opts.userAgent,
    });
    const getCreds = getCredentialsByURI.bind(null, opts.rawConfig);
    return {
        tarball: fetchFromTarball.bind(null, {
            fetchFromRemoteTarball: fetchFromRemoteTarball.bind(null, {
                download,
                getCredentialsByURI: mem((registry) => getCreds(registry)),
                offline: opts.offline,
            }),
        }),
    };
}
exports.default = default_1;
function fetchFromTarball(ctx, cafs, resolution, opts) {
    if (resolution.tarball.startsWith('file:')) {
        const tarball = resolvePath(opts.lockfileDir, resolution.tarball.slice(5));
        return fetchFromLocalTarball(cafs, tarball, {
            integrity: resolution.integrity,
            manifest: opts.manifest,
        });
    }
    return ctx.fetchFromRemoteTarball(cafs, resolution, opts);
}
const isAbsolutePath = /^[/]|^[A-Za-z]:/;
function resolvePath(where, spec) {
    if (isAbsolutePath.test(spec))
        return spec;
    return path.resolve(where, spec);
}
async function fetchFromRemoteTarball(ctx, cafs, dist, opts) {
    try {
        return await fetchFromLocalTarball(cafs, opts.cachedTarballLocation, {
            integrity: dist.integrity,
            manifest: opts.manifest,
        });
    }
    catch (err) {
        // ignore errors for missing files or broken/partial archives
        switch (err.code) {
            case 'Z_BUF_ERROR':
                if (ctx.offline) {
                    throw new error_1.default('CORRUPTED_TARBALL', `The cached tarball at "${opts.cachedTarballLocation}" is corrupted. Cannot redownload it as offline mode was requested.`);
                }
                logger_1.globalWarn(`Redownloading corrupted cached tarball: ${opts.cachedTarballLocation}`);
                break;
            case 'EINTEGRITY':
                if (ctx.offline) {
                    throw new error_1.default('BAD_TARBALL_CHECKSUM', `The cached tarball at "${opts.cachedTarballLocation}" did not pass the integrity check. Cannot redownload it as offline mode was requested.`);
                }
                logger_1.globalWarn(`The cached tarball at "${opts.cachedTarballLocation}" did not pass the integrity check. Redownloading.`);
                break;
            case 'ENOENT':
                if (ctx.offline) {
                    throw new error_1.default('NO_OFFLINE_TARBALL', `Could not find ${opts.cachedTarballLocation} in local registry mirror`);
                }
                break;
            default:
                throw err;
        }
        const auth = dist.registry ? ctx.getCredentialsByURI(dist.registry) : undefined;
        return ctx.download(dist.tarball, {
            auth,
            cafs,
            integrity: dist.integrity,
            manifest: opts.manifest,
            onProgress: opts.onProgress,
            onStart: opts.onStart,
            registry: dist.registry,
        });
    }
}
async function fetchFromLocalTarball(cafs, tarball, opts) {
    try {
        const tarballStream = fs.createReadStream(tarball);
        const [fetchResult] = (await Promise.all([
            cafs.addFilesFromTarball(tarballStream, opts.manifest),
            opts.integrity && ssri.checkStream(tarballStream, opts.integrity),
        ]));
        return { filesIndex: fetchResult };
    }
    catch (err) {
        err.attempts = 1;
        err.resource = tarball;
        throw err;
    }
}
//# sourceMappingURL=index.js.map